import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
from typing import Literal, Tuple, Any, Optional
from pathlib import Path


# TODO: Exception Handling.
# TODO: add multicolored-line from matplotlib.

class DataVisualizer:
    @staticmethod
    def plot(
            data: pd.DataFrame,
            plot_type: Literal[
                "scatter", "line", "bar", "hist", "box", "violin", "heatmap", "pairplot", "jointplot", "kde", "swarm", "lmplot"],
            x: Optional[str] = None,
            y: Optional[str] = None,
            title: str = "",
            figsize: Tuple[float, float] = (10, 6),
            theme: Optional[str] = None,
            palette: Optional[str] = None,
            yticks: Optional[list] = None,
            **kwargs: Any
    ) -> None:
        """
        Create various types of plots based on the provided data.

        Args:
            data (pd.DataFrame): The dataset to visualize.
            plot_type (str): The type of plot to create.
            x (str, optional): The column name for x-axis.
            y (str, optional): The column name for y-axis.
            title (str): The title of the plot.
            figsize (Tuple[float, float]): The size of the figure in inches.
            theme (str, optional): The seaborn theme to use.
            palette (str, optional): The color palette to use.
            yticks (list, optional): adjust the scale of y-axis.
            **kwargs: Additional keyword arguments for the specific plot type.

        Returns:
            None
        """
        if theme:
            sns.set_theme(theme)

        if palette:
            sns.set_palette(palette)

        fig, ax = plt.subplots(figsize=figsize)

        if plot_type == "scatter":
            sns.scatterplot(data=data, x=x, y=y, ax=ax, **kwargs)
        elif plot_type == "line":
            sns.lineplot(data=data, x=x, y=y, ax=ax, **kwargs)
        elif plot_type == "bar":
            sns.barplot(data=data, x=x, y=y, ax=ax, **kwargs)
        elif plot_type == "hist":
            sns.histplot(data=data, x=x, y=y, ax=ax, **kwargs)
        elif plot_type == "box":
            sns.boxplot(data=data, x=x, y=y, ax=ax, **kwargs)
        elif plot_type == "violin":
            sns.violinplot(data=data, x=x, y=y, ax=ax, **kwargs)
        elif plot_type == "heatmap":
            sns.heatmap(data=data, ax=ax, annot=kwargs.pop('annot', True), **kwargs)
        elif plot_type == "pairplot":
            sns.pairplot(data=data, **kwargs)
            plt.suptitle(title, y=1.02)
            return
        elif plot_type == "jointplot":
            sns.jointplot(data=data, x=x, y=y, **kwargs)
            plt.suptitle(title, y=1.02)
            return
        elif plot_type == "kde":
            sns.kdeplot(data=data, x=x, y=y, ax=ax, **kwargs)
        elif plot_type == "swarm":
            sns.swarmplot(data=data, x=x, y=y, ax=ax, **kwargs)
        elif plot_type == "lmplot":
            sns.lmplot(data=data, x=x, y=y, **kwargs)
            plt.title(title)
            return
        else:
            raise ValueError(f"Unsupported plot type: {plot_type}")

        plt.title(title)
        if x:
            plt.xlabel(x)
        if y:
            plt.ylabel(y)

        # Rotate x-axis labels if they're too long
        if x and len(data[x].unique()) > 10:
            plt.xticks(rotation=45, ha='right')

        if yticks:
            ax.set_yticks(yticks)

        plt.tight_layout()
        plt.show()

    @staticmethod
    def multi_plot(
            data: pd.DataFrame,
            plot_specs: list,
            title: str = "",
            figsize: Tuple[float, float] = (15, 10),
            theme: Optional[str] = None,
            palette: Optional[str] = None,
            yticks: Optional[list] = None
    ) -> None:
        """
        Create multiple plots in a single figure.

        Args:
            data (pd.DataFrame): The dataset to visualize.
            plot_specs (list): A list of dictionaries, each specifying a subplot.
                Each dict should contain 'type', 'x', 'y', and any additional kwargs.
            title (str): The main title of the figure.
            figsize (Tuple[float, float]): The size of the figure in inches.
            theme (str, optional): The seaborn theme to use.
            palette (str, optional): The color palette to use.
            yticks (list, optional): adjust the scale of y-axis.

        Returns:
            None
        """
        if theme:
            sns.set_theme(theme)

        if palette:
            sns.set_palette(palette)

        n_plots = len(plot_specs)
        rows = (n_plots + 1) // 2  # Calculate number of rows (2 plots per row)
        fig, axes = plt.subplots(rows, 2, figsize=figsize)
        axes = axes.flatten()  # Flatten axes array for easy indexing

        for i, plot_spec in enumerate(plot_specs):
            plot_type = plot_spec.pop('type')
            x = plot_spec.pop('x', None)
            y = plot_spec.pop('y', None)

            if plot_type == "scatter":
                sns.scatterplot(data=data, x=x, y=y, ax=axes[i], **plot_spec)
            elif plot_type == "line":
                sns.lineplot(data=data, x=x, y=y, ax=axes[i], **plot_spec)
            elif plot_type == "bar":
                sns.barplot(data=data, x=x, y=y, ax=axes[i], **plot_spec)
            elif plot_type == "hist":
                sns.histplot(data=data, x=x, y=y, ax=axes[i], **plot_spec)
            elif plot_type == "box":
                sns.boxplot(data=data, x=x, y=y, ax=axes[i], **plot_spec)
            elif plot_type == "violin":
                sns.violinplot(data=data, x=x, y=y, ax=axes[i], **plot_spec)
            elif plot_type == "kde":
                sns.kdeplot(data=data, x=x, y=y, ax=axes[i], **plot_spec)
            elif plot_type == "swarm":
                sns.swarmplot(data=data, x=x, y=y, ax=axes[i], **plot_spec)
            else:
                raise ValueError(f"Unsupported plot type in multi_plot: {plot_type}")

            axes[i].set_title(plot_spec.get('subtitle', ''))
            if x:
                axes[i].set_xlabel(x)
            if y:
                axes[i].set_ylabel(y)

        # Remove any unused subplots
        for i in range(n_plots, len(axes)):
            fig.delaxes(axes[i])

        if yticks:
            axes.set_yticks(yticks)

        plt.suptitle(title, fontsize=16)
        plt.tight_layout()
        plt.show()

    @staticmethod
    def plot_melted_comparison(
            data: pd.DataFrame,
            x: str,
            y_vars: list,
            plot_type: Literal["scatter", "line", "bar"],
            var_name: str = "variable",
            value_name: str = "value",
            title: str = "",
            figsize: Tuple[float, float] = (10, 6),
            theme: Optional[str] = None,
            palette: Optional[str] = None,
            yticks: Optional[list] = None,
            **kwargs: Any
    ) -> None:
        """
        Plot a comparison of multiple y-variables on the same plot by melting the data.

        Args:
            data (pd.DataFrame): The dataset to visualize.
            x (str): The column name for x-axis.
            y_vars (list): List of column names to compare on y-axis.
            var_name (str): Name for the melted variable column (default: 'variable').
            value_name (str): Name for the melted value column (default: 'value').
            title (str): The title of the plot.
            figsize (Tuple[float, float]): The size of the figure in inches.
            theme (str, optional): The seaborn theme to use.
            palette (str, optional): The color palette to use.
            yticks (list, optional): adjust the scale of y-axis.
            **kwargs: Additional keyword arguments for the line plot.

        Returns:
            None
        """
        if theme:
            sns.set_theme(theme)

        if palette:
            sns.set_palette(palette)

        # Melt the data to have sales and avg_sales in a single column
        melted_data = pd.melt(data, id_vars=[x], value_vars=y_vars, var_name=var_name, value_name=value_name)

        fig, ax = plt.subplots(figsize=figsize)

        if plot_type == "line":
            sns.lineplot(data=melted_data, x=x, y=value_name, hue=var_name, ax=ax, **kwargs)
        elif plot_type == "bar":
            sns.barplot(data=melted_data, x=x, y=value_name, hue=var_name, ax=ax, **kwargs)
        elif plot_type == "scatter":
            sns.scatterplot(data=melted_data, x=x, y=value_name, hue=var_name, ax=ax, **kwargs)

        plt.title(title)
        plt.xlabel(x)
        plt.ylabel(value_name)

        # Rotate x-axis labels if they're too long
        if len(data[x].unique()) > 10:
            plt.xticks(rotation=45, ha='right')

        if yticks:
            ax.set_yticks(yticks)

        plt.tight_layout()
        plt.show()

    def save_fig(self,
                 fig_id: int,
                 tight_layout: bool = True,
                 fig_extension: str = "png",
                 location: str = "images",
                 resolution: int = 300):
        """Saves the image in a location."""
        path = Path() / location / f"{fig_id}.{fig_extension}"
        path.mkdir(parents=True, exist_ok=True)
        if tight_layout:
            plt.tight_layout()
        plt.savefig(path, format=fig_extension, dpi=resolution)




from openai import OpenAI, APIError, APIConnectionError, OpenAIError, RateLimitError
from dotenv import load_dotenv
from sqthon.util import (
    database_schema,
    format_database_schema,
    make_dataframe_json_serializable)
import os
from sqlalchemy import Engine, text
import json
import pandas as pd
from typing import final
from tenacity import retry, wait_random_exponential, stop_after_attempt


class LLM:
    def __init__(self, model: str, connection: Engine):
        load_dotenv()
        self.model = model
        self.connection = connection
        self.client = OpenAI(api_key=os.getenv("OPENAI_API_KEY"))
        self.db_schema = database_schema(self.connection)
        self.messages = [
            {
                "role": "developer",
                "content": """
                            You are an SQL expert, known for generating efficient and optimized SQL queries. 
                            Your primary role is to generate optimized SQL queries.
                            When presenting query results:
                            - Use proper Markdown syntax for formatting.
                            - Use `#` and `##` headers for sections and `---` for horizontal breaks.
                            - Present numerical data with proper formatting (e.g., thousands separator,
                              decimal precision).
                            - Use Markdown tables for structured data and limit output to the first 20 rows unless the
                              user specifies otherwise.
                            - Highlight key insights or observations where applicable.

                            When presenting schemas or query plans:
                            - Format them clearly using Markdown code blocks (` ```sql ` for SQL or ` ``` ` 
                              for plain text).
                            - Ensure the schema is organized and easy to understand.

                            Additional Guidelines:
                            - If the user's input is ambiguous or incomplete, ask clarifying questions.
                        """,
            }
        ]
        self.tools = [
            {
                "type": "function",
                "function": {
                    "name": "ask_db",
                    "description": f"""Use this function to answer user questions about database tables using 
                            {self.connection.engine.dialect.name} syntax. Input should be a fully formed SQL query.""",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "query": {
                                "type": "string",
                                "description": f"""
                                    SQL query extracting info to answer the the user's question.
                                    SQL should be written using this database schema:
                                    {format_database_schema(self.db_schema)}
                                    The query should be returned in plain text, not in JSON.
                                """,
                            }
                        },
                        "required": ["query"],
                    },
                },
            }
        ]

        self.last_query_result = None
        self.max_messages = 30

    def trim_chat(self):
        total_msg = 0
        for _ in self.messages:
            total_msg += 1
            if total_msg > self.max_messages:
                self.messages.pop(1)
                self.messages.pop(1)

    @final
    @retry(wait=wait_random_exponential(multiplier=1, max=10), stop=stop_after_attempt(3))
    def get_response(self):
        try:
            return self.client.chat.completions.create(
                model=self.model, messages=self.messages,
                tools=self.tools, tool_choice="auto",
                temperature=0.3
            )
        except (APIError, APIConnectionError, OpenAIError, RateLimitError) as e:
            print(f"Error occurred: {e}")


    def execute_fn(self, show_query: bool = False):
        """
        Checks for model responses and executes ask_db method.
        Parameters:
            - show_query (bool): Show the generated query if True.
        """

        try:
            response = self.get_response()
            response_msg = response.choices[0].message
            self.messages.append(response_msg)

            if response_msg.tool_calls:
                tool_call = response_msg.tool_calls[0]
                tool_call_id = tool_call.id
                function_name = tool_call.function.name

                if function_name == "ask_db":
                    query = json.loads(tool_call.function.arguments)["query"]

                    if show_query:
                        print(query)

                    result = self.ask_db(query)
                    results_json = make_dataframe_json_serializable(result)

                    if len(json.dumps(results_json)) > 100:
                        results_json = make_dataframe_json_serializable(result[:20])
                        results_json.append(
                            {"summary": f"{len(result)} rows fetched. Showing the first 30 rows only."}
                        )

                    self.messages.append(
                        {
                            "role": "tool",
                            "tool_call_id": tool_call_id,
                            "name": function_name,
                            "content": json.dumps(results_json),
                        }
                    )
                    try:
                        final_response = self.client.chat.completions.create(
                            model=self.model, messages=self.messages, temperature=0.3
                        )
                        return final_response.choices[0].message.content
                    except RateLimitError as e:
                        print(f"Rate limit exceeded: {e}")

                else:
                    raise ValueError(f"Unknown function: {function_name}")
            else:
                return response_msg.content

        except Exception as e:
            raise Exception(f"Error in execute_fn: {str(e)}")

    def ask_db(self, query: str) -> pd.DataFrame:
        """Function to query  databases with a provided SQL query."""
        try:
            result = pd.read_sql_query(text(query), self.connection)
            self.last_query_result = result
            return result
        except Exception as e:
            raise Exception(f"Error executing query: {e}")



import pandas as pd
from sqlalchemy import (
    MetaData, Column, Table, Integer, Float, Numeric, String, Text, Boolean,
    DateTime, Date, Time, JSON, ARRAY, LargeBinary, Interval, Engine, inspect
)
from sqlalchemy.exc import ResourceClosedError
from typing import List
import json
import tiktoken


def map_dtype_to_sqlalchemy(dtype):
    """
    Comprehensive mapping of pandas dtypes to SQLAlchemy types.

    Args:
        dtype: pandas dtype object or string

    Returns:
        SQLAlchemy type object with nullable property set
    """
    dtype_str = str(dtype).lower()

    # Numeric types
    if 'int' in dtype_str:
        return Integer()
    # elif 'int64' in dtype_str or 'uint64' in dtype_str:
    #     return BigInteger()
    elif 'float' in dtype_str:
        return Float(precision=53)
    elif 'decimal' in dtype_str:
        return Numeric(precision=38, scale=10)

    # Boolean type
    elif 'bool' in dtype_str:
        return Boolean()

    # DateTime types
    elif 'datetime64[ns]' in dtype_str:
        return DateTime()
    elif 'timedelta' in dtype_str:
        return Interval()
    elif 'date' in dtype_str:
        return Date()
    elif 'time' in dtype_str:
        return Time()

    # String types
    elif 'string' in dtype_str or 'object' in dtype_str:
        return String(length=255)
    elif 'category' in dtype_str:
        return String(length=64)

    # Complex types
    elif 'complex' in dtype_str:
        return String(length=100)

    # JSON type (for dictionary/list columns)
    elif isinstance(dtype, pd.api.types.CategoricalDtype):
        return String(length=64)
    elif dtype_str == 'object' and pd.api.types.is_dict_like(pd.Series([{}, None])):
        return JSON()

    # Array types
    elif 'array' in dtype_str:
        return ARRAY(String)

    # Binary types
    elif 'bytes' in dtype_str:
        return LargeBinary()

    # Default to Text for any unhandled types
    else:
        return Text()


def create_table(path: str,
                 table_name: str,
                 engine: Engine,
                 key: bool = False) -> Table:
    """Reads a csv file and creates a table.

    Parameters:
        - path (str): Path to the csv file.
        - table_name (str): Name of the table you want to create.
        - engine (Engine): Sqlalchemy's engine.
        - key (bool): Whether to keep primary key or not.

    Returns:
        - Table
    """
    metadata_obj = MetaData()
    df = pd.read_csv(path, nrows=5)
    # TODO: make the column tuple.
    columns = []
    if key:
        columns.append(Column("id", primary_key=True, autoincrement=True))
    for col, dtype in df.dtypes.items():
        sqlalchemy_type = map_dtype_to_sqlalchemy(dtype)
        columns.append(Column(str(col), sqlalchemy_type, nullable=True))

    table = Table(table_name, metadata_obj, *columns)

    metadata_obj.create_all(engine)

    return table


def to_datetime(df, column):
    """
    Convert a column in a DataFrame to datetime type.

    Args:
    df (pd.DataFrame): The DataFrame containing the column to convert.
    column (str): The name of the column to convert to datetime.

    Returns:
    pd.DataFrame: The DataFrame with the specified column converted to datetime.
    """
    try:
        df[column] = pd.to_datetime(df[column])
        return df
    except Exception as e:
        print(f"Error converting column {column} to datetime: {e}")
        return df


def table_keys(table_name: str, connection: Engine) -> dict:
    """Return primary and foreign keys of the table."""
    inspector = inspect(connection)
    if table_name not in inspector.get_table_names():
        raise ValueError(f"{table_name} doesn't exist.")

    primary_keys = inspector.get_pk_constraint(table_name).get("constrained_columns", [])
    foreign_keys = [
        {
            "column": fk["constrained_columns"][0],
            "referred_table": fk["referred_table"],
            "referred_column": fk["referred_columns"][0]
        } for fk in inspector.get_foreign_keys(table_name)
    ]

    return {"primary_keys": primary_keys, "foreign_keys": foreign_keys}


def indexes(table: str, connection: Engine) -> list:
    return inspect(connection).get_indexes(table_name=table)


def get_table_schema(table: str, connection: Engine) -> List:
    """Returns schema of the specific table."""
    try:
        inspector = inspect(connection)
        return [{"column_name": col["name"], "column_data_type": col["type"]} for col in inspector.get_columns(table)]
    except ResourceClosedError as e:
        print(f"An error occurred: {e}")


def tables(connection: Engine) -> List:
    """Returns a list with available table names."""
    try:
        return inspect(connection).get_table_names()
    except ResourceClosedError as e:
        print(f"An error occurred: {e}")


def database_schema(connection: Engine) -> List:
    """Returns schema of the database."""
    return [
        {
            "table_name": table,
            "keys": table_keys(table, connection),
            "column_names_with_dtypes": [
                {"name": column["column_name"], "data_type": column["column_data_type"]}
                for column in get_table_schema(table, connection)
            ]
        }
        for table in tables(connection)
    ]


def date_dimension(connection: Engine,
                   year_start: str,
                   year_end: str,
                   freq: str = 'D') -> pd.DataFrame:
    """
    Creates a date dimension table using Pandas to generate date series.

    Parameters:
    -----------
    connection : Engine
        SQLAlchemy database connection engine
    table_name : str
        Name of the table to create
    year_start : str
        Start date of the date series (e.g., '2000-01-01')
    year_end : str
        End date of the date series (e.g., '2010-12-31')
    freq : str, optional
        Pandas frequency alias (default is 'D' for daily)
        Common values:
        - 'D': Daily
        - 'B': Business daily
        - 'W': Weekly
        - 'M': Monthly
        - 'Q': Quarterly
        - 'Y': Yearly
    """
    date_series = pd.date_range(start=year_start, end=year_end, freq=freq)

    return pd.DataFrame({
        'date': date_series,
        'date_key': date_series.strftime('%Y%m%d').astype(int),
        'day_of_month': date_series.day,
        'day_of_year': date_series.dayofyear,
        'day_of_week': date_series.dayofweek + 1,  # Pandas uses 0-6, actual is  1-7
        'day_name': date_series.strftime('%A'),
        'day_short_name': date_series.strftime('%a'),
        'week_number': date_series.isocalendar().week,
        'week_of_month': ((date_series.day - 1) // 7) + 1,
        'week': date_series - pd.to_timedelta(date_series.dayofweek, unit='D'),
        'month_number': date_series.month,
        'month_name': date_series.strftime('%B'),
        'month_short_name': date_series.strftime('%b'),
        'first_day_of_month': date_series.to_period('M').strftime('%Y-%m-%d'),
        'last_day_of_month': date_series.to_period('M').strftime('%Y-%m-%d'),
        'quarter_number': date_series.quarter,
        'quarter_name': 'Q' + date_series.quarter.astype(str),
        'first_day_of_quarter': date_series.to_period('Q').strftime('%Y-%m-%d'),
        'last_day_of_quarter': date_series.to_period('Q').strftime('%Y-%m-%d'),
        'year': date_series.year,
        'decade': (date_series.year // 10) * 10,
        'century': (date_series.year // 100) * 100
    })


def make_dataframe_json_serializable(df: pd.DataFrame):
    """
            Converts a DataFrame to a JSON-serializable format dynamically by handling:
            - NaN/None values
            - Datetime columns
            - Object columns with complex types (lists, dicts, etc.)
        """
    df = df.fillna("")  # replacing Nan with empty strings.
    # 2. Handle datetime columns dynamically.
    for col in df.select_dtypes(include=["datetime", "datetimetz"]).columns:
        df[col] = df[col].dt.strftime("%Y-%m-%d")  # Convert datetime to string format

    # 3. Handle object or complex types (e.g., dict, list, tuple, etc.)
    def handle_complex_types(value):
        if isinstance(value, (dict, list, tuple)):  # If value is dict or list or tuple, convert to JSON string
            return json.dumps(value)
        return value

    for col in df.select_dtypes(include=["object"]).columns:
        df[col] = df[col].apply(handle_complex_types)

    # 4. Convert to JSON-serializable dictionary
    return df.to_dict(orient="records")


def format_database_schema(db_schema: List):
    """
    Format database schema into a readable string representation
    """
    return "\n".join(
        [
            f"Table Name: {table['table_name']}\n"
            f"Primary keys-> {", ".join(item for item in table['keys']['primary_keys'])}\n"
            f"Foreign keys-> {", ".join(f'{k}: {v}' for item in table["keys"]["foreign_keys"] for k, v in item.items())}\n"
            f"Column Names with Dtypes->\n{'\n'.join(f"{item['name']}: {str(item['data_type'])}" for item in table['column_names_with_dtypes'])}\n"
            for table in db_schema
        ]
    )


def num_tokens_from_messages(messages: List, model: str):
    """Returns the number of tokens used by a list of messages."""
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using o200k_base encoding.")
        encoding = tiktoken.get_encoding("o200k_base")
    if model in {
        "gpt-3.5-turbo-0125",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
        "gpt-4o-mini-2024-07-18",
        "gpt-4o-2024-08-06",
    }:
        tokens_per_msg = 3
        tokens_per_name = 1

    elif "gpt-3.5-turbo" in model:
        print(
            "Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0125."
        )
        return num_tokens_from_messages(messages, model="gpt-3.5-turbo-0125")
    elif "gpt-4o-mini" in model:
        print(
            "Warning: gpt-4o-mini may update over time. Returning num tokens assuming gpt-4o-mini-2024-07-18."
        )
        return num_tokens_from_messages(
            messages, model="gpt-4o-mini-2024-07-18"
        )
    elif "gpt-4o" in model:
        print(
            "Warning: gpt-4o and gpt-4o-mini may update over time. Returning num tokens assuming gpt-4o-2024-08-06."
        )
        return num_tokens_from_messages(messages, model="gpt-4o-2024-08-06")
    elif "gpt-4" in model:
        print(
            "Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613."
        )
        return num_tokens_from_messages(messages, model="gpt-4-0613")
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model}."""
        )
    num_tokens = 0
    for msg in messages:
        num_tokens += tokens_per_msg
        for key, value in msg.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is  primed with <|start|>assistant<|message|>
    return num_tokens


def count_tokens_for_tools(functions: List, messages: List, model: str):
    # Initialize function settings to 0
    func_init = 0
    prop_init = 0
    prop_key = 0
    enum_init = 0
    enum_item = 0
    func_end = 0

    if model in ["gpt-4o", "gpt-4o-mini"]:
        # Set function settings for the above models
        func_init = 7
        prop_init = 3
        prop_key = 3
        enum_init = -3
        enum_item = 3
        func_end = 12
    elif model in ["gpt-3.5-turbo", "gpt-4"]:
        # Set function settings for the above models
        func_init = 10
        prop_init = 3
        prop_key = 3
        enum_init = -3
        enum_item = 3
        func_end = 12
    else:
        raise NotImplementedError(
            f"""num_tokens_for_tools() is not implemented for model {model}."""
        )
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        print("Warning: model not found. Using o200k_base encoding.")
        encoding = tiktoken.get_encoding("o200k_base")

    func_token_count = 0
    if len(functions) > 0:
        for f in functions:
            func_token_count += func_init
            function = f["function"]
            f_name = function["name"]
            f_desc = function["description"]
            if f_desc.endswith("."):
                f_desc = f_desc[:-1]
            line = f_name + ":" + f_desc
            func_token_count += len(encoding.encode(line))
            if len(function["parameters"]["properties"]) > 0:
                func_token_count += prop_init
                for key in list(function["parameters"]["properties"].keys()):
                    func_token_count += prop_key
                    p_name = key
                    p_type = function["parameters"]["properties"][key]["type"]
                    p_desc = function["parameters"]["properties"][key][
                        "description"
                    ]
                    if "enum" in function["parameters"]["properties"][key].keys():
                        func_token_count += enum_init
                        for item in function["parameters"]["properties"][key][
                            "enum"
                        ]:
                            func_token_count += enum_item
                            func_token_count += len(encoding.encode(item))
                    if p_desc.endswith("."):
                        p_desc = p_desc[:-1]
                    line = f"{p_name}:{p_type}:{p_desc}"
                    func_token_count += len(encoding.encode(line))
        func_token_count += func_end

    messages_token_count = num_tokens_from_messages(messages, model)
    return messages_token_count + func_token_count



import traceback
from sqthon.connection import DatabaseConnector
from typing import final
from sqlalchemy import text
from sqlalchemy.exc import OperationalError, ProgrammingError
from typing import Literal
from sqthon.db_context import DatabaseContext


class Sqthon:
    def __init__(
            self, dialect: str, user: str, host: str, service_instance_name: str = None
    ):
        self.dialect = dialect
        self.user = user
        self.host = host
        self.connect_db = DatabaseConnector(
            dialect=self.dialect,
            user=self.user,
            host=self.host,
            service_instance_name=service_instance_name,
        )
        self.connections = {}

    def server_infile_status(self) -> bool:
        """
        Checks for global infile status in the server.
        Returns:
            True: if it's on.
            False: if it's off.
        """
        try:
            with self.connect_db.server_level_engine().connect() as conn:
                global_infile = conn.execute(
                    text("SHOW GLOBAL VARIABLES LIKE 'local_infile';")
                ).fetchone()[1]
        except (OperationalError, ProgrammingError) as e:
            print(f"An error occurred: {e}")

        return global_infile.lower() == "on"

    def global_infile_mode(self, mode: Literal["on", "off"]):
        """
        Enable or disable global infile.
        Parameters:
            mode (str): 'on' or 'off'.
        """
        try:
            with self.connect_db.server_level_engine().connect() as conn:
                if mode.lower() == "on":
                    conn.execute(text("SET GLOBAL local_infile = 1"))
                elif mode.lower() == "off":
                    conn.execute(text("SET GLOBAL local_infile = 0"))
                else:
                    raise ValueError("Invalid mode. Expected 'on' opr 'off'.")
        except (OperationalError, ProgrammingError) as e:
            print(f"An error occurred: {e}")

    def session_infile_mode(self, mode: Literal["on", "off"]):
        """Enable or disable session infile."""
        try:
            with self.connect_db.server_level_engine().connect() as conn:
                if mode.lower() == "on":
                    conn.execute(text("SET SESSION local_infile = 1"))
                elif mode.lower() == "off":
                    conn.execute(text("SET SESSION local_infile = 0"))
                else:
                    raise ValueError("Invalid mode. Expected 'on' or 'off'.")
        except (OperationalError, ProgrammingError) as e:
            print(f"An error occurred: {e}")

    @final
    def file_permission(self, access: Literal["grant", "revoke"]):
        """Give or remove access to a user."""
        try:
            with self.connect_db.server_level_engine().connect() as conn:
                if access.lower() == "grant":
                    conn.execute(
                        text(f"GRANT FILE ON *.* TO '{self.user}'@'{self.host}'")
                    )
                elif access.lower() == "revoke":
                    conn.execute(
                        text(f"REVOKE FILE ON *.* TO '{self.user}'@'{self.host}'")
                    )
                else:
                    raise ValueError("Invalid mode. Expected 'grant' or 'revoke'")
        except OperationalError:
            print("Failed to connect.")
            traceback.print_exc()

    def create_database(self, database: str):
        try:
            with self.connect_db.server_level_engine().connect() as connection:
                try:
                    if self.dialect.lower() == "mysql":
                        connection.execute(
                            text(f"CREATE DATABASE IF NOT EXISTS {database};")
                        )
                    elif self.dialect.lower() == "postgresql":
                        if not connection.execute(
                                text(
                                    f"SELECT 1 FROM pg_database WHERE datname = '{database}'"
                                )
                        ).scalar():
                            connection.execute(text(f"CREATE DATABASE {database}"))
                except ProgrammingError as e:
                    print(f"Programming error: {e}")
        except OperationalError:
            print("Failed to connect.")

    def show_dbs(self):
        try:
            with self.connect_db.server_level_engine().connect() as connection:
                if self.dialect.lower() == "mysql":
                    dbs = connection.execute(text(f"SHOW DATABASES;")).fetchall()
                elif self.dialect.lower() == "postgresql":
                    dbs = connection.execute((text("SELECT datname FROM pg_database;")))
        except OperationalError:
            print("Server is not running.")

        return [db[0] for db in dbs]

    def drop_db(self, database: str):
        with self.connect_db.server_level_engine().connect() as connection:
            connection.execute(text(f"DROP DATABASE {database};"))

    @final
    def connect_to_database(self, database: str = None, local_infile: bool = False, use_llm: bool = False,
                            model: str = None):
        """Connects to specific database."""
        try:
            connection = self.connect_db.connect(
                database=database, local_infile=local_infile
            )
            self.connections[database] = DatabaseContext(
                database=database, connection=connection, llm=use_llm, model_name=model
            )
        except Exception as e:
            print(f"Error connecting to database {database}: {e}")
            traceback.print_exc()

        return self.connections[database]

    @final
    def show_connections(self):
        return [key for key in self.connect_db.connections]

    def disconnect_database(self, database):
        self.connect_db.disconnect(database)



from sqlalchemy import text, Engine
from sqlalchemy.exc import (
    OperationalError,
    DataError,
    ProgrammingError,
    IntegrityError,
    ResourceClosedError,
)
from typing import Literal, Callable, final
from sqthon.util import create_table
from sqthon.util import (
    get_table_schema,
    tables,
    date_dimension,
    indexes,
    database_schema,
)
import os
import pandas as pd
from sqthon.llm import LLM
from sqthon.data_visualizer import DataVisualizer
from rich import print as rprint


@final
class DatabaseContext:
    """Context-specific sub-instance for a specific database."""

    def __init__(self,
                 database: str,
                 connection: Engine,
                 llm: bool = False,
                 model_name: str = None,
                 ):
        self.database = database
        self.connection = connection
        self.visualizer = DataVisualizer()
        if llm:
            self.llm = LLM(model=model_name, connection=self.connection)

    def get_tables(self) -> list:
        """Returns the names of available tables"""
        return tables(self.connection)

    def check_indexes(self, table: str) -> list:
        """Check indexes for the table."""
        return indexes(table=table, connection=self.connection)

    def table_schema(self, table: str) -> list:
        return get_table_schema(table=table, connection=self.connection)

    def get_database_schema(self) -> list:
        """Returns the schema of the database."""
        return database_schema(self.connection)

    def drop_table(self, table: str) -> None:
        """Drops a table from the database."""
        self.connection.execute(text(f"DROP TABLE {table}"))


    def ask(
            self, prompt: str, as_df: bool = False, display_query: bool = True
    ) -> str | pd.DataFrame:
        """
        Ask a question about the database.
        Args:
            prompt (str): The question to ask
            as_df (bool): If True, returns the raw DataFrame instead of formatted response
            display_query (bool): If True, prints the generated SQL query

        Returns:
            Union[str, pd.DataFrame]: Either formatted response or DataFrame based on as_df parameter
        """
        try:
            self.llm.messages.append({"role": "user", "content": prompt})
            self.llm.trim_chat()
            result = self.llm.execute_fn(show_query=display_query)

            # if show_token_usage and token_count is not None:
            #     print(f"Token Usage: {token_count} tokens used.")

            if as_df and self.llm.last_query_result is not None:
                return self.llm.last_query_result
            else:
                rprint(result)
                # return result

        except Exception as e:
            raise Exception(f"Error in ask method: {str(e)}")

    def generate_date_series(
            self,
            table: str,
            start_year: str,
            end_year: str,
            frequency: str = "D",
            if_exists: Literal["replace", "fail", "append"] = "fail",
            insert_method: Literal["multi"] | Callable | None = None,
            index: bool = True,
    ):
        """
        Creates and populates a date dimension table from a specific year upto a specific year.

        Parameters:
        -----------
        table_name : str
            Name of the table.
        start_year : str
            start date of the date series.
        end_year : str
            end year of the date series.
        freq : str, optional
            Pandas frequency alias (default is 'D' for daily)
            Common values:
            - 'D': Daily
            - 'B': Business daily
            - 'W': Weekly
            - 'M': Monthly
            - 'Q': Quarterly
            - 'Y': Yearly

        if_exists : str
            How to behave if table already exists.
        insert_method : {None, ‘multi’, callable}, optional
            Controls the SQL insertion clause used:
            None : Uses standard SQL INSERT clause (one per row).
            ‘multi’: Pass multiple values in a single INSERT clause.


        """

        df = date_dimension(
            connection=self.connection,
            year_start=start_year,
            year_end=end_year,
            freq=frequency,
        )

        df.to_sql(
            name=table,
            con=self.connection,
            if_exists=if_exists,
            method=insert_method,
            index=index,
        )

    def import_csv_to_mysqldb(
            self, csv_path: str, table: str, terminated_by: str = "\n"
    ):
        """
        Imports a CSV file into a MySQL database table with flexible import options.

        This method provides two primary import mechanisms:
        1. Using pandas: Leverages pandas' DataFrame to_sql method for CSV import
        2. Using direct SQL LOAD DATA INFILE for more performance-oriented imports

        Parameters:
        -----------
        csv_path : str
            The absolute or relative path to the source CSV file.
            Must be a valid, accessible file path.

        table : str
            The name of the target MySQL table for data import.
            If the table doesn't exist, it will be automatically created.

        terminated_by : str, optional
            Line termination character for CSV parsing.
            Defaults to newline ("\n").

        Returns:
        --------
        None
            Imports data directly into the specified MySQL table.

        Raises:
        -------
        FileNotFoundError
            If the specified CSV file does not exist.

        """
        if not os.path.exists(csv_path):
            raise FileNotFoundError(f"CSV file not found: {csv_path}")

        try:
            table = create_table(
                engine=self.connection, table_name=table, path=csv_path
            )
            columns = [col.name for col in table.columns]
            col_name_clause = ", ".join([f"`{name.strip()}`" for name in columns])
            query = text(
                f"""
            LOAD DATA LOCAL INFILE '{csv_path}'
            INTO TABLE {table}
            FIELDS TERMINATED BY ','
            LINES TERMINATED BY '{terminated_by}'
            IGNORE 1 ROWS
            ({col_name_clause})
            """
            )

            self.connection.execute(query)
            self.connection.commit()

        except (
                OperationalError,
                ProgrammingError,
                ResourceClosedError,
                IntegrityError,
                DataError,
        ) as e:
            self.connection.rollback()
            raise RuntimeError(f"Error importing CSV: {e}")

    def run_query(
            self,
            query: str,
            plot_type: (
                    Literal[
                        "scatter",
                        "line",
                        "bar",
                        "hist",
                        "box",
                        "violin",
                        "heatmap",
                        "pairplot",
                        "jointplot",
                        "kde",
                        "swarm",
                        "lmplot",
                    ]
                    | None
            ) = None,
            visualize: bool = False,
            x=None,
            y=None,
            title=None,
            **kwargs,
    ) -> pd.DataFrame | None:
        """
        Executes a SQL query and optionally visualizes the result.

        Parameters:
            - query (str): The SQL query to be executed.
            - visualize (bool, optional): If True, the result will be visualized. Default is False.
            - plot_type (str, optional): The type of plot to create if visualize is True.
                Must be one of 'scatter', 'line', or 'bar'. Default is None.
            - x (str, optional): The column name to be used for the x-axis in the plot. Required if visualize is True.
            - y (str, optional): The column name to be used for the y-axis in the plot. Required if visualize is True.
            - title (str, optional): The title for the plot. Required if visualize is True.
            - **kwargs: Additional keyword arguments passed to the plotting function.

        Returns:
            - result (Object): The result of the SQL query execution.

        Raises:
            - ValueError: If visualize is True but plot_type, x, y, or title are not provided.
        """

        try:
            result = pd.read_sql_query(text(query), self.connection)
            if visualize:
                if not all([plot_type, x, y]):
                    raise ValueError(
                        "For visualization, please provide plot_type, x, y."
                    )
                self.visualizer.plot(result, plot_type, x, y, title, **kwargs)

            return result

        except ProgrammingError as e:
            print(f"Programming error: {e}")
        except Exception as e:
            print(f"Error executing query: {e}")
            return None



import os
from dotenv import load_dotenv
from sqlalchemy import create_engine, URL
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError, ArgumentError
from typing import final


# TODO: Dialects to be added: SQlite ✅, Oracle, Microsoft SQL Server.
# TODO: Isolation_level for connections.
# TODO: support for encrypted connections.
# TODO: Exception handling


class DatabaseConnector:
    """
    A class to manage database connections for MySQL or PostgreSQL.

    This class facilitates creating connections to databases, handling credentials securely,
    and managing engine instances for both database-level and server-level operations.

    Attributes:
        dialect (str): Database dialect, such as 'mysql' or 'postgresql'.
        user (str): Username for the database.
        host (str): Database host address.
        service_instance_name (str, optional): Name of the service instance.
        engines (dict): Dictionary storing created SQLAlchemy engines.
        connections (dict): Dictionary storing active connections.
    """

    def __init__(
        self, dialect: str, user: str, host: str, service_instance_name: str = None
    ):
        """
        Initializes the DatabaseConnector instance with specified connection parameters.

        Parameters:
            dialect (str): The database dialect ('mysql' or 'postgresql').
            user (str): The database user for authentication.
            host (str): The host address of the database.
            service_instance_name (str, optional): An identifier for the database service instance, if any.

        Important:
            To enhance security, avoid hardcoding sensitive information like passwords in your code.
            Store the database password in a `.env` file using the format '<username>password=yourpassword'
            and load it using `dotenv`. This will be accessed as an environment variable.
        """
        load_dotenv()
        self.dialect = dialect
        self.user = user
        self.host = host
        self.service_instance_name = service_instance_name
        self.engines = {}
        self.connections = {}

        if self.dialect.lower() == "mysql":
            self.driver = "pymysql"
        elif self.dialect.lower() == "postgresql":
            self.driver = "psycopg2"
        elif self.dialect.lower() == "sqlite":
            self.driver = None

    @final
    def _create_engine(
        self, database: str, local_infile: bool, pool_size: int, max_overflow: int
    ) -> Engine:

        try:

            if self.dialect.lower() == "sqlite":
                return create_engine(f"sqlite:///{database}")

            # TODO: if more than one same username exists then password fetching gonna give problems.
            password = os.getenv(f"{self.user}password")
            if not password:
                raise ValueError(
                    f"Password for user '{self.user}' not found in environment variables."
                )

            url_object = URL.create(
                f"{self.dialect}+{self.driver}",
                username=self.user,
                password=password,
                host=self.host,
                database=database,
            )

            connection_args = {
                "local_infile": local_infile,
            }

            return create_engine(
                url_object,
                connect_args=connection_args,
                pool_size=pool_size,
                max_overflow=max_overflow,
                pool_recycle=3600,
            )
        except ArgumentError as ae:
            print(f"Incorrect arguments: {ae}")

    @final
    def server_level_engine(
        self, database: str = None, local_infile: bool = False
    ) -> Engine:
        """Use this engine for server level one-time operation.
        It's good to use this method using context manager.

        Parameters:
        database (str, optional): The database name. If not provided, the engine will be created without a specific
                                  database.
        local_infile (bool): whether to enable or disable local data infile.

        Returns:
        Engine: The SQLAlchemy engine for server-level operations.
        """
        try:
            if database:
                url_object = URL.create(
                    f"{self.dialect}+{self.driver}",
                    username=self.user,
                    password=os.getenv(f"{self.user}password"),
                    host=self.host,
                    database=database,
                )
            else:
                url_object = URL.create(
                    f"{self.dialect}+{self.driver}",
                    username=self.user,
                    password=os.getenv(f"{self.user}password"),
                    host=self.host,
                )

            args = {"local_infile": local_infile}

            return create_engine(url_object, connect_args=args)
        except ArgumentError as e:
            print(f"Incorrect arguments: {e}")

    @final
    def connect(
        self,
        database: str,
        local_infile: bool,
        pool_size: int = 20,
        max_overflow: int = 10,
    ):
        if database not in self.connections or self.connections[database].closed:
            try:
                if database not in self.engines:
                    self.engines[database] = self._create_engine(
                        database, local_infile, pool_size, max_overflow
                    )
                self.connections[database] = self.engines[database].connect()
            except OperationalError:
                from sqthon.services import start_service

                print(f"Looks like {self.dialect} server instance is not running.")
                print("Trying to start the server...")
                try:
                    if self.service_instance_name is None:
                        self.service_instance_name = input(
                            "Enter the server instance name: "
                        )
                    start_service(self.service_instance_name)
                    self.connections[database] = self.engines[database].connect()
                    return self.connections[database]
                except Exception:
                    raise RuntimeError(
                        f"Not able established the server! Try to start manually."
                    )

        return self.connections[database]

    @final
    def disconnect(self, database):
        if database in self.connections and self.connections[database] is not None:
            try:
                self.connections[database].close()
            except Exception as e:
                print(f"Error closing the connection: {e}")
            finally:
                del self.connections[database]

    @final
    def dispose_engine(self, database: str):
        """Permanently disposes an engine."""
        if database in self.engines and self.engines[database] is not None:
            self.engines[database].dispose()



import requests
from requests.exceptions import RequestException
import subprocess
import os
import traceback
import win32event, win32process
from win32com.shell.shell import ShellExecuteEx
import win32com.shell.shellcon as shellcon


def is_admin():
    if os.name == "nt":
        import ctypes

        try:
            return ctypes.windll.shell32.IsUserAnAdmin()
        except:
            traceback.print_exc()
            return False


def _runAsAdmin(service: str, action: str):
    cmd = "net"
    params = f"{action} {service}"
    execute_cmd = ShellExecuteEx(
        lpVerb="runas",
        lpFile=cmd,
        lpParameters=params,
        fMask=shellcon.SEE_MASK_NOCLOSEPROCESS,
    )

    if execute_cmd:
        handle = execute_cmd["hProcess"]
        win32event.WaitForSingleObject(handle, win32event.INFINITE)

        exit_code = win32process.GetExitCodeProcess(handle)
        print(f"Process exited with code: {exit_code}")

        if exit_code == 0:
            if action == "start":
                print(f"{service} server instance started successfully.")
            elif action == "stop":
                print(f"{service} server instance stopped successfully.")
            return True
        else:
            if action == "start":
                print(f"Failed to start the {service} server. Exit code: {exit_code}")
            elif action == "stop":
                print(f"Failed to stop the {service} server. Exit code: {exit_code}")
    else:
        if action == "start":
            print(f"Failed to start {service} server instance.")
        elif action == "stop":
            print(f"Failed to stop the {service} server instance.")


def is_ollama_running(url: str = "http://localhost:11434") -> bool:
    """
    Check if the ollama service is running.

    Args:
    url (str): The URL of the Ollama service. Defaults to "http://localhost:11434".

    Returns:
    bool: True if the Ollama service is running, False otherwise.
    """

    try:
        response = requests.get(f"{url}/api/version", timeout=5)
        return response.status_code == 200
    except RequestException:
        return False


def is_service_running(service_instance_name: str) -> bool:
    check_service_status = subprocess.run(
        ["sc", "query", service_instance_name],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    return "RUNNING" in check_service_status.stdout


def start_ollama_service():
    try:
        subprocess.run(["ollama", "serve"], check=True)
        print("Ollama service started successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Failed to start the ollama service: {e}")


def start_service(service_name: str):
    """Starts a service instance with elevated privileges if needed.
    Parameters:
          service_name (str): Name of the server to start.
    """
    if is_service_running(service_name):
        print(f"{service_name} server instance is already running.")
        return
    if 0 != is_admin():
        try:
            subprocess.run(["sc", "start", service_name], check=True)
            print(f"{service_name} server instance has started.")
        except subprocess.CalledProcessError as e:
            print(f" Failed to execute command: {e.returncode}")
            print(f"Error message: {e.stderr}")
    else:
        try:
            _runAsAdmin(service_name, "start")
            return True
        except Exception as e:
            print(f"Error occurred: {e}")


def stop_service(service_name: str):
    if not is_service_running(service_name):
        print(f"{service_name} server is already stopped.")
        return
    if 0 != is_admin():
        try:
            subprocess.run(["sc", "stop", service_name], check=True)
            print(f"{service_name} server instance has started.")
        except subprocess.CalledProcessError as e:
            print(f" Failed to execute command: {e.returncode}")
            print(f"Error message: {e.stderr}")
    else:
        try:
            _runAsAdmin(service_name, "stop")
        except Exception as e:
            print(f"{e}")



class ServiceManagementError(Exception):
    """Base exception class for service management operation."""
    pass


class ServiceStartError(ServiceManagementError):
    """Raised when failed to start a service."""

    def __init__(self,
                 service_name: str,
                 msg: str = None,
                 exit_code: int = None):
        ...


class ServiceStopError(ServiceManagementError):
    """Raised when failed to stop a service."""

    def __init__(self):
        ...


class AdminPrivilegeError(ServiceManagementError):
    """Raised when admin privilege required but not available."""

    def __init__(self):
        ...


class ServiceConnectionError(ServiceManagementError):
    """Raised when unable to connect to a service."""

    def __init__(self):
        ...
